/*
 *   teamd_dbus_common.h - Teamd dbus api common things
 *   Copyright (C) 2013-2015 Jiri Pirko <jiri@resnulli.us>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _TEAMD_DBUS_COMMON_H_
#define _TEAMD_DBUS_COMMON_H_

#define TEAMD_DBUS_SERVICE	"org.libteam.teamd"
#define TEAMD_DBUS_IFACE	"org.libteam.teamd"
#define TEAMD_DBUS_PATH		"/org/libteam/teamd"

#endif /* _TEAMD_DBUS_COMMON_H_ */
